from ontobio.io.assocparser import AssocParser
from ontobio.io import assocparser
from ontobio.io.assocparser import ENTITY, EXTENSION, ANNOTATION

class HpoaParser(AssocParser):
    """
    Parser for HPOA format

    http://human-phenotype-ontology.github.io/documentation.html#annot

    Note that there are similarities with Gaf format, so we inherit from GafParser, and override
    """

    def __init__(self,config=assocparser.AssocParserConfig()):
        """
        Arguments:
        ---------

        config : a AssocParserConfig object
        """

        self.config = config
        self.report = assocparser.Report(config=self.config)

    def skim(self, file):
        file = self._ensure_file(file)
        tuples = []
        for line in file:
            if line.startswith("!"):
                continue
            vals = line.split("\t")
            if len(vals) < 14:
                logging.error("Unexpected number of vals: {}.".format(vals))

            negated, relation, _ = self._parse_qualifier(vals[3], vals[8])

            # never include NOTs in a skim
            if negated:
                continue
            if self._is_exclude_relation(relation):
                continue
            id = self._pair_to_id(vals[0], vals[1])
            if not self._validate_id(id, line, context=ENTITY):
                continue
            n = vals[2]
            t = vals[4]
            tuples.append( (id,n,t) )
        return tuples

    def parse_line(self, line):
        """
        Parses a single line of a HPOA file

        Return a tuple `(processed_line, associations)`. Typically
        there will be a single association, but in some cases there
        may be none (invalid line) or multiple (disjunctive clause in
        annotation extensions)

        Note: most applications will only need to call this directly if they require fine-grained control of parsing. For most purposes,
        :method:`parse_file` can be used over the whole file

        Arguments
        ---------
        line : str
            A single tab-seperated line from a GPAD file

        """
        config = self.config

        parsed = super().validate_line(line)
        if parsed:
            return parsed

        if self.is_header(line):
            return assocparser.ParseResult(line, [], False)

        # http://human-phenotype-ontology.github.io/documentation.html#annot
        vals = line.split("\t")
        if len(vals) != 14:
            self.report.error(line, assocparser.Report.WRONG_NUMBER_OF_COLUMNS, "",
                msg="There were {columns} columns found in this line, and there should be 14".format(columns=len(vals)))
            return assocparser.ParseResult(line, [], True)

        [db,
         db_object_id,
         db_object_symbol,
         qualifier,
         hpoid,
         reference,
         evidence,
         onset,
         frequency,
         withfrom,
         aspect,
         db_object_synonym,
         date,
         assigned_by] = vals

        # hardcode this, as HPOA is currently human-only
        taxon = 'NCBITaxon:9606'
        split_line = assocparser.SplitLine(line=line, values=vals, taxon=taxon)


        # hardcode this, as HPOA is currently disease-only
        db_object_type = 'disease'

        ## --
        ## db + db_object_id. CARD=1
        ## --
        id = self._pair_to_id(db, db_object_id)
        if not self._validate_id(id, split_line, context=ENTITY):
            return assocparser.ParseResult(line, [], True)

        if not self._validate_id(hpoid, split_line, context=ANNOTATION):
            return assocparser.ParseResult(line, [], True)

        valid_hpoid = self._validate_ontology_class_id(hpoid, split_line)
        if valid_hpoid == None:
            return assocparser.ParseResult(line, [], True)
        hpoid = valid_hpoid

        # validation
        #self._validate_symbol(db_object_symbol, line)

        #TODO: HPOA has different date styles
        #date = self._normalize_gaf_date(date, line)

        # Example use case: mapping from OMIM to Orphanet
        if config.entity_map is not None:
            id = self.map_id(id, config.entity_map)
            toks = id.split(":")
            db = toks[0]
            db_object_id = toks[1:]
            vals[1] = db_object_id

        ## --
        ## end of line re-processing
        ## --
        # regenerate line post-mapping
        line = "\t".join(vals)

        ## --
        ## db_object_synonym CARD=0..*
        ## --
        synonyms = db_object_synonym.split("|")
        if db_object_synonym == "":
            synonyms = []


        ## --
        ## qualifier
        ## --
        ## we generate both qualifier and relation field
        relation = None
        qualifiers = qualifier.split("|")
        if qualifier == '':
            qualifiers = []
        negated =  'NOT' in qualifiers
        other_qualifiers = [q for q in qualifiers if q != 'NOT']

        ## CURRENTLY NOT USED
        if len(other_qualifiers) > 0:
            relation = other_qualifiers[0]
        else:
            if aspect == 'O':
                relation = 'has_phenotype'
            elif aspect == 'I':
                relation = 'has_inheritance'
            elif aspect == 'M':
                relation = 'mortality'
            elif aspect == 'C':
                relation = 'has_onset'
            else:
                relation = None

        # With/From
        withfroms = self.validate_pipe_separated_ids(withfrom, split_line, empty_allowed=True, extra_delims=",")
        if withfroms == None:
            # Reporting occurs in above function call
            return assocparser.ParseResult(line, [], True)

        ## --
        ## hpoid
        ## --
        object = {'id':hpoid,
                  'taxon': taxon}

        # construct subject dict
        subject = {
            'id':id,
            'label':db_object_symbol,
            'type': db_object_type,
            'synonyms': synonyms,
            'taxon': {
                'id': taxon
            }
        }

        ## --
        ## evidence
        ## reference
        ## withfrom
        ## --
        evidence = {
            'type': evidence,
            'has_supporting_reference': reference.split("; "),
            'with_support_from': withfroms
        }

        ## Construct main return dict
        assoc = {
            'source_line': line,
            'subject': subject,
            'object': object,
            'negated': negated,
            'qualifiers': qualifiers,
            'relation': {
                'id': relation
            },
            'interacting_taxon': None,
            'evidence': evidence,
            'provided_by': assigned_by,
            'date': date,

        }

        return assocparser.ParseResult(line, [assoc], False)

    def is_header(self, line):
        return line.startswith("!")



"""
Mapping between native ontology model and rdflib
"""
from ontobio.ontol import Ontology, Synonym
from prefixcommons.curie_util import contract_uri, expand_uri, get_prefixes
from rdflib import Namespace
from rdflib import BNode
from rdflib import Literal
from rdflib import URIRef
from rdflib.namespace import RDFS
from rdflib.namespace import OWL
import networkx
import logging

def rdfgraph_to_ontol(rg):
    """
    Return an Ontology object from an rdflib graph object

    Status: Incomplete
    """
    digraph = networkx.MultiDiGraph()
    from rdflib.namespace import RDF
    label_map = {}
    for c in rg.subjects(RDF.type, OWL.Class):
        cid = contract_uri_wrap(c)
        logging.info("C={}".format(cid))
        for lit in rg.objects(c, RDFS.label):
            label_map[cid] = lit.value
            digraph.add_node(cid, label=lit.value)
        for s in rg.objects(c, RDFS.subClassOf):
            # todo - blank nodes
            sid = contract_uri_wrap(s)
            digraph.add_edge(sid, cid, pred='subClassOf')

    logging.info("G={}".format(digraph))
    payload = {
        'graph': digraph,
        #'xref_graph': xref_graph,
        #'graphdoc': obographdoc,
        #'logical_definitions': logical_definitions
    }
            
    ont = Ontology(handle='wd', payload=payload)
    return ont



                
        
def contract_uri_wrap(uri):
    curies = contract_uri(uri)
    if len(curies) > 0:
        return curies[0]
    else:
        return uri



import io
import json

from ontobio.io import assocparser
from ontobio.io import gafparser

from ontobio.io import assocwriter

def test_no_colon_in_id():
    parser = gafparser.GafParser()
    valid = parser._validate_id("FOOBAR", assocparser.SplitLine("", [""] * 17, "taxon:foo"))

    assert not valid
    assert len(parser.report.messages) == 1
    assert parser.report.messages[0]["level"] == assocparser.Report.ERROR

def test_empty_id():
    parser = gafparser.GafParser()
    valid = parser._validate_id("", assocparser.SplitLine("", [""] * 17, "taxon:foo"))
    assert not valid
    assert len(parser.report.messages) == 1
    assert parser.report.messages[0]["level"] == assocparser.Report.ERROR

def test_pipe_in_id():
    parser = gafparser.GafParser()
    valid = parser._validate_id("F|OO:123", assocparser.SplitLine("", [""] * 17, "taxon:foo"))

    assert not valid
    assert len(parser.report.messages) == 1
    assert parser.report.messages[0]["level"] == assocparser.Report.ERROR

def test_bad_character_in_id():
    parser = gafparser.GafParser()
    valid = parser._validate_id("FOO:1&23", assocparser.SplitLine("", [""] * 17, "taxon:foo"))

    assert not valid
    assert len(parser.report.messages) == 1
    assert parser.report.messages[0]["level"] == assocparser.Report.ERROR

def test_empty_post_colon():
    parser = gafparser.GafParser()
    valid = parser._validate_id("FOO:", assocparser.SplitLine("", [""] * 17, "taxon:foo"))

    assert not valid
    assert len(parser.report.messages) == 1
    assert parser.report.messages[0]["level"] == assocparser.Report.ERROR

def test_empty_pre_colon():
    parser = gafparser.GafParser()
    valid = parser._validate_id(":123", assocparser.SplitLine("", [""] * 17, "taxon:foo"))

    assert not valid
    assert len(parser.report.messages) == 1
    assert parser.report.messages[0]["level"] == assocparser.Report.ERROR

def test_validate_with_allowed_ids():
    parser = gafparser.GafParser()
    valid = parser._validate_id("FOO:123", assocparser.SplitLine("", [""]*17, "taxon:foo"), allowed_ids=["FOO"])
    assert valid

def test_validate_with_disallowed_id():
    parser = gafparser.GafParser()
    valid = parser._validate_id("FOO:123", assocparser.SplitLine("", [""]*17, "taxon:foo"), allowed_ids=["BAR"])
    assert len(parser.report.messages) == 1
    assert parser.report.messages[0]["level"] == assocparser.Report.WARNING

def test_validate_pipe_separated():
    parser = gafparser.GafParser()
    ids = parser.validate_pipe_separated_ids("PMID:12345", assocparser.SplitLine("", [""] * 17, "taxon:foo"))
    assert set(ids) == set(["PMID:12345"])

    ids = parser.validate_pipe_separated_ids("PMID:12345|PMID:11111", assocparser.SplitLine("", [""] * 17, "taxon:foo"))
    assert set(ids) == set(["PMID:12345", "PMID:11111"])

def test_validate_pipe_separated_with_bad_ids():
    parser = gafparser.GafParser()
    ids = parser.validate_pipe_separated_ids("PMID:123[2]|PMID:11111", assocparser.SplitLine("", [""] * 17, "taxon:foo"))

    assert ids == None

    ids = parser.validate_pipe_separated_ids("PMID:123[2]", assocparser.SplitLine("", [""] * 17, "taxon:foo"))
    assert ids == None

def test_validate_pipe_separated_empty_allowed():
    parser = gafparser.GafParser()
    ids = parser.validate_pipe_separated_ids("", assocparser.SplitLine("", [""] * 17, "taxon:foo"), empty_allowed=True)

    assert ids == []

def test_validate_pipe_with_additional_delims():
    parser = gafparser.GafParser()
    ids = parser.validate_pipe_separated_ids("F:123,B:234|B:111", assocparser.SplitLine("", [""] * 17, "taxon:foo"), extra_delims=",")

    assert set(ids) == set(["F:123", "B:234", "B:111"])

    result = parser.parse_line("PomBase\tSPAC25B8.17\typf1\t\tGO:1990578\tGO_REF:0000024\tISO\tUniProtKB:Q9CXD9|ensembl:ENSMUSP00000038569,PMID:11111\tC\tintramembrane aspartyl protease of the perinuclear ER membrane Ypf1 (predicted)\tppp81\tprotein\ttaxon:4896\t20150305\tPomBase\t\t")
    assert set(result.associations[0]["evidence"]["with_support_from"]) == set(["UniProtKB:Q9CXD9", "ensembl:ENSMUSP00000038569", "PMID:11111"])

def test_doi_id():
    parser = gafparser.GafParser()
    valid = parser._validate_id("DOI:10.1007/BF00127499", assocparser.SplitLine("", [""]*17, "taxon:foo"))
    assert valid



from ontobio.config import Config, get_config, set_config, reset_config
import logging

def _endpoint(e):
    return "URL: {} OBJ: {}".format(e.url, str(e))

def test_cfg():
    logging.info("TEST")
    cfg = get_config()

    logging.info("GOT: {}".format(cfg))
    print(str(cfg))
    assert True
    #cfg = get_config()
    print("Main solr search: {}".format(_endpoint(cfg.solr_search)))
    print("Main solr assocs: {}".format(_endpoint(cfg.solr_assocs)))
    print("Pre-loaded ontologies: {}".format(cfg.ontologies))
    
    set_config('tests/resources/test-config.yaml')
    cfg = get_config()

    # reset
    reset_config()
    
    assert cfg.solr_assocs.url == "https://example.org"

    assert cfg.get_category_class('anatomy') == 'PO:0025131'

    
    assert cfg.ontologies[0].id == "magic1234"    
    assert cfg.ontologies[0].handle == "pato"    
    assert cfg.ontologies[0].pre_load == True
    



from ontobio.io import assocparser
from ontobio.io import entityparser
from ontobio.io import entitywriter
from ontobio.io.assocparser import ENTITY, EXTENSION, ANNOTATION

import logging

class GpadParser(assocparser.AssocParser):
    """
    Parser for GO GPAD Format

    https://github.com/geneontology/go-annotation/blob/master/specs/gpad-gpi-1_2.md
    """


    def __init__(self, config=assocparser.AssocParserConfig(), group="unknown", dataset="unknown"):
        """
        Arguments:
        ---------

        config : a AssocParserConfig object
        """
        self.config = config
        self.report = assocparser.Report(config=self.config, group="unknown", dataset="unknown")
        self.gpi = None
        if self.config.gpi_authority_path is not None:
            self.gpi = dict()
            parser = entityparser.GpiParser()
            with open(self.config.gpi_authority_path) as gpi_f:
                entities = parser.parse(file=gpi_f)
                for entity in entities:
                    self.gpi[entity["id"]] = {
                        "symbol": entity["label"],
                        "name": entity["full_name"],
                        "synonyms": entitywriter.stringify(entity["synonyms"]),
                        "type": entity["type"]
                    }
                print("Loaded {} entities from {}".format(len(self.gpi.keys()), self.config.gpi_authority_path))


    def skim(self, file):
        file = self._ensure_file(file)
        tuples = []
        for line in file:
            if line.startswith("!"):
                continue
            vals = line.split("\t")
            if len(vals) != 12:
                logging.error("Unexpected number of columns: {}. GPAD should have 12.".format(vals))
            rel = vals[2]

            negated, relation, _ = self._parse_qualifier(vals[2], None)

            # never include NOTs in a skim
            if negated:
                continue
            if self._is_exclude_relation(relation):
                continue


            id = self._pair_to_id(vals[0], vals[1])
            if not self._validate_id(id, line, context=ENTITY):
                continue
            t = vals[3]
            tuples.append( (id,None,t) )
        return tuples

    def parse_line(self, line):
        """Parses a single line of a GPAD.

        Return a tuple `(processed_line, associations)`. Typically
        there will be a single association, but in some cases there
        may be none (invalid line) or multiple (disjunctive clause in
        annotation extensions)

        Note: most applications will only need to call this directly if they require fine-grained control of parsing. For most purposes,
        :method:`parse_file` can be used over the whole file

        Arguments
        ---------
        line : str
            A single tab-seperated line from a GPAD file

        """
        parsed = super().validate_line(line)
        if parsed:
            return parsed

        if self.is_header(line):
            return assocparser.ParseResult(line, [{ "header": True, "line": line.strip() }], False)

        vals = [el.strip() for el in line.split("\t")]
        if len(vals) < 10 or len(vals) > 12:
            self.report.error(line, assocparser.Report.WRONG_NUMBER_OF_COLUMNS, "",
                msg="There were {columns} columns found in this line, and there should be between 10 and 12".format(columns=len(vals)))
            return assocparser.ParseResult(line, [], True)

        if len(vals) < 12:
            vals += [""] * (12 - len(vals))

        [db,
         db_object_id,
         qualifier,
         goid,
         reference,
         evidence,
         withfrom,
         interacting_taxon_id,
         date,
         assigned_by,
         annotation_xp,
         annotation_properties] = vals

        split_line = assocparser.SplitLine(line=line, values=vals, taxon="")

        id = self._pair_to_id(db, db_object_id)
        if not self._validate_id(id, split_line, context=ENTITY):
            return assocparser.ParseResult(line, [], True)

        if not self._validate_id(goid, split_line, context=ANNOTATION):
            return assocparser.ParseResult(line, [], True)

        valid_goid = self._validate_ontology_class_id(goid, split_line)
        if valid_goid == None:
            return assocparser.ParseResult(line, [], True)
        goid = valid_goid

        date = self._normalize_gaf_date(date, split_line)

        if reference == "":
            self.report.error(line, Report.INVALID_ID, "EMPTY", "reference column 6 is empty")
            return assocparser.ParseResult(line, [], True)

        self._validate_id(evidence, split_line)

        interacting_taxon = None if interacting_taxon_id == "" else interacting_taxon_id
        if interacting_taxon != None:
            interacting_taxon = self._taxon_id(interacting_taxon_id, split_line)
            if interacting_taxon == None:
                self.report.error(line, assocparser.Report.INVALID_TAXON, interacting_taxon_id,
                                    msg="Taxon ID is invalid")
                return assocparser.ParseResult(line, [], True)

        #TODO: ecomap is currently one-way only
        #ecomap = self.config.ecomap
        #if ecomap != None:
        #    if ecomap.ecoclass_to_coderef(evidence) == (None,None):
        #        self.report.error(line, Report.UNKNOWN_EVIDENCE_CLASS, evidence,
        #                          msg="Expecting a known ECO class ID")

        ## --
        ## qualifier
        ## --
        negated, relation, other_qualifiers = self._parse_qualifier(qualifier, None)


        # Reference Column
        references = self.validate_pipe_separated_ids(reference, split_line)
        if references == None:
            # Reporting occurs in above function call
            return assocparser.ParseResult(line, [], True)

        # With/From
        withfroms = self.validate_pipe_separated_ids(withfrom, split_line, empty_allowed=True, extra_delims=",")
        if withfroms == None:
            # Reporting occurs in above function call
            return assocparser.ParseResult(line, [], True)


        ## --
        ## parse annotation extension
        ## See appending in http://doi.org/10.1186/1471-2105-15-155
        ## --
        object_or_exprs = self._parse_full_extension_expression(annotation_xp, line=split_line)

        subject_symbol = id
        subject_fullname = id
        subject_synonyms = []
        if self.gpi is not None:
            gp = self.gpi.get(id, {})
            if gp is not {}:
                subject_symbol = gp["symbol"]
                subject_fullname = gp["name"]
                subject_synonyms = gp["synonyms"].split("|")

        assoc = {
            'source_line': line,
            'subject': {
                'id': id,
                'label': subject_symbol,
                'fullname': subject_fullname,
                'synonyms': subject_synonyms,
                'taxon': {
                    'id': interacting_taxon
                },
            },
            'object': {
                'id':goid
            },
            'negated': negated,
            'relation': {
                'id': relation
            },
            'interacting_taxon': interacting_taxon,
            'evidence': {
                'type': evidence,
                'with_support_from': withfroms,
                'has_supporting_reference': references
            },
            'subject_extensions': [],
            'object_extensions': {},
            'aspect': self.compute_aspect(goid),
            'provided_by': assigned_by,
            'date': date,
        }
        if len(other_qualifiers) > 0:
            assoc['qualifiers'] = other_qualifiers
        if object_or_exprs is not None and len(object_or_exprs) > 0:
            assoc['object_extensions'] = {'union_of': object_or_exprs}


        return assocparser.ParseResult(line, [assoc], False)

    def is_header(self, line):
        return line.startswith("!")



from ontobio import OntologyFactory
from ontobio import GraphRenderer
import logging

HAS_PART = 'BFO:0000051'
PART_OF = 'BFO:0000050'

QUALITY = 'PATO:0000001'
PLOIDY = 'PATO:0001374'
EUPLOID = 'PATO:0001393'
SHAPE = 'PATO:0000052'
Y_SHAPED = 'PATO:0001201'
PENTAPLOID = 'PATO:0001383'
SWOLLEN = 'PATO:0001851'
DILATED = 'PATO:0001571'
INCREASED_SIZE = 'PATO:0000586'
PROTRUDING = 'PATO:0001598'
MORPHOLOGY = 'PATO:0000051'
ABSENT = 'PATO:0000462'
CONICAL = 'PATO:0002021'
APOPTOTIC = 'PATO:0000638'

def test_remote_sparql_pato():
    """
    Load ontology from remote SPARQL endpoint
    """
    factory = OntologyFactory()
    print("Creating ont")
    ont = factory.create('pato')

    ont_id = ont.id
    assert 'pato' in ont_id.lower()
    
    ploidy = ont.node(PLOIDY)
    print("PLOIDY: {}".format(ploidy))
    assert ont.label(PLOIDY) == 'ploidy'

    # exact match
    search_results = ont.search('shape')
    print("SEARCH (exact): {}".format(search_results))
    assert [SHAPE] == search_results

    # implicit regexp
    search_results = ont.search('%shape%')
    print("SEARCH (re, implicit): {}".format(search_results))
    assert SHAPE in search_results
    assert len(search_results)>10

    # explicit regexp
    search_results = ont.search('.*shape.*', is_regex=True)
    print("SEARCH (re, explicit): {}".format(search_results))
    assert SHAPE in search_results
    assert len(search_results)>10
    
    # syns
    syn = 'cone-shaped'
    search_results = ont.search(syn, synonyms=False)
    print("SEARCH (no syns): {}".format(search_results))
    assert [] == search_results
    #search_results = ont.search(syn, synonyms=True)
    #print("SEARCH (with syns): {}".format(search_results))
    #assert [CONICAL] == search_results
    
    num_nodes = 0
    for n in ont.nodes():
        num_nodes = num_nodes+1
    assert num_nodes > 100

    ancs = ont.ancestors(PLOIDY)
    print("ANCS ploidy (all): {}".format(ancs))
    assert QUALITY in ancs
    assert PENTAPLOID not in ancs

    ancs = ont.ancestors(PLOIDY, relations=['subClassOf'])
    print("ANCS ploidy (subClassOf): {}".format(ancs))
    assert QUALITY in ancs
    assert PENTAPLOID not in ancs

    # this is a non-use case
    ancs = ont.ancestors(SWOLLEN, relations=[HAS_PART])
    print("ANCS swollen (has_part): {}".format(ancs))
    assert INCREASED_SIZE in ancs
    assert PROTRUDING in ancs
    assert len(ancs) == 2

    ancs = ont.ancestors(SWOLLEN, relations=['subClassOf'])
    print("ANCS swollen (has_part): {}".format(ancs))
    assert MORPHOLOGY in ancs
    assert QUALITY in ancs
    assert PROTRUDING not in ancs
    
    decs = ont.descendants(PLOIDY)
    print("DECS ploidy (all): {}".format(decs))
    assert QUALITY not in decs
    assert EUPLOID in decs
    assert PENTAPLOID in decs

    # this is a non-use case
    ancs = ont.descendants(INCREASED_SIZE, relations=[HAS_PART])
    print("ANCS increased size (has part): {}".format(ancs))
    assert SWOLLEN in ancs
    assert len(ancs) == 1

    subsets = ont.subsets()
    print("SUBSETS: {}".format(subsets))

    slim = ont.extract_subset('absent_slim')
    print("SLIM: {}".format(slim))
    assert ABSENT in slim
    assert QUALITY not in slim

    syns = ont.synonyms(INCREASED_SIZE)
    print("SYNS: {}".format(syns))
    syn_vals = [syn.val for syn in syns]
    assert 'big' in syn_vals
    [bigsyn] = [syn for syn in syns if syn.val=='big']
    # TODO xrefs
    assert not bigsyn.exact_or_label()
    assert bigsyn.scope() == 'RELATED'

    defn = ont.text_definition(INCREASED_SIZE)
    assert defn is not None

    assert ont.is_obsolete('PATO:0000000') == True
    
    w = GraphRenderer.create('obo')
    w.write_subgraph(ont, [INCREASED_SIZE])

    assert len(ont.xref_graph.nodes()) > 1
    assert 'GO:0006915' in ont.xrefs(APOPTOTIC)
    w = GraphRenderer.create('obo')
    w.write_subgraph(ont, [APOPTOTIC])
    
def test_dynamic_query():
    """
    Dynamic query
    """
    factory = OntologyFactory()
    print("Creating ont")
    ont = factory.create('pato')

    ids = ont.sparql(body="{?x rdfs:subClassOf+ "+SHAPE+"}",
                     inject_prefixes = ont.prefixes(),
                     single_column=True)
    assert Y_SHAPED in ids
    assert ABSENT not in ids

def test_subontology():
    """
    subontology
    """
    factory = OntologyFactory()
    print("Creating ont")
    ont = factory.create('go')
    assert ont.is_obsolete('GO:0000267') == True
    print("ONT NODES: {}".format(ont.nodes()))
    subont = ont.subontology(relations=['subClassOf'])
    PERM = 'GO:1990578'
    print("NODES: {}".format(subont.nodes()))
    ancs = subont.ancestors(PERM, reflexive=True)
    print(str(ancs))
    for a in ancs:
        print(" ANC: {} '{}'".format(a,subont.label(a)))
    assert len(ancs) > 0

    assert subont.is_obsolete('GO:0000267') == True
    
    w = GraphRenderer.create('tree')
    w.write_subgraph(ont, ancs)

    # TODO: sub-ontology does not create
    # full metadata
    w = GraphRenderer.create('obo')
    w.write_subgraph(ont, ancs)
    



"""
Various classes for rendering of ontology graphs
"""

import networkx as nx
from networkx.drawing.nx_pydot import write_dot
import tempfile
import os
import subprocess
import json
import logging

class GraphRendererConfig():
    """
    configuration parameters
    """
    def __init__(self,
                 relsymbolmap=None,
                 show_text_definition=False):
        self.relsymbolmap = relsymbolmap
        self.show_text_definition = show_text_definition
        if self.relsymbolmap is None:
            self.relsymbolmap = {
                     'subClassOf': '%',
                     'subPropertyOf': '%',
                     'BFO:0000050': '<',
                     'RO:0002202': '~',
            }
        

class GraphRenderer():
    """
    base class for writing networkx graphs
    """
    def __init__(self,
                 outfile=None,
                 config=None,
                 **args):
        self.outfile = outfile
        if config is None:
            config = GraphRendererConfig()
        self.config = config
        
    
    def render(self, ontol, **args):
        """
        Render a `ontology` object
        """
        pass

    def write(self, ontol, **args):
        """
        Write a `ontology` object
        """
        s = self.render(ontol, **args)
        if self.outfile is None:
            print(s)
        else:
            f = open(self.outfile, 'w')
            f.write(s)
            f.close()

    def render_subgraph(self, ontol, nodes, **args):
        """
        Render a `ontology` object after inducing a subgraph
        """
        subont = ontol.subontology(nodes, **args)
        return self.render(subont, **args)
    
    def write_subgraph(self, ontol, nodes, **args):
        """
        Write a `ontology` object after inducing a subgraph
        """
        subont = ontol.subontology(nodes, **args)
        self.write(subont, **args)

    def render_relation(self, r, **args):
        """
        Render an object property
        """
        if r is None:
            return "."
        m = self.config.relsymbolmap
        if r in m:
            return m[r]
        return r
    
    def render_noderef(self, ontol, n, query_ids=None, **args):
        """
        Render a node object
        """
        if query_ids is None:
            query_ids = []
        marker = ""
        if n in query_ids:
            marker = " * "
        label = ontol.label(n)
        s = None
        if label is not None:
            s = '{} ! {}{}'.format(n,
                                   label,
                                   marker)
        else:
            s = str(n)
        if self.config.show_text_definition:
            td = ontol.text_definition(n)
            if td:
                s += ' "{}"'.format(td.val)
        return s
        
    @staticmethod
    def create(fmt):
        """
        Creates a GraphRenderer
        """
        w = None
        if fmt == 'tree':
            w = AsciiTreeGraphRenderer()
        elif fmt == 'dot':
            w = DotGraphRenderer(image_format='dot')
        elif fmt == 'png':
            w = DotGraphRenderer(image_format='png')
        elif fmt == 'ndot':
            w = NativeDotGraphRenderer()
        elif fmt == 'obo':
            w = OboFormatGraphRenderer()
        elif fmt == 'obog':
            w = OboJsonGraphRenderer()
        else:
            w = SimpleListGraphRenderer()
        return w
        
class NativeDotGraphRenderer(GraphRenderer):
    """
    writes as dot (graphviz format) files
    """
    def __init__(self, **args):
        super().__init__(**args)

    def render(self, ontol, **args):
        g = ontol.get_graph()
        _, fn = tempfile.mkstemp(suffix='dot')
        write_dot(g, fn)
        f = open(fn, 'r')
        s = f.read()
        f.close()
        return s

    def write(self, ontol, **args):
        g = ontol.get_graph()        
        fn = self.outfile
        if fn is None:
            _, fn = tempfile.mkstemp(suffix='dot')
        write_dot(g, fn)
        if self.outfile is None:
            f = open(fn, 'r')
            print(f.read())
            f.close()
            os.remove(fn)

class DotGraphRenderer(GraphRenderer):
    """
    uses js lib to generate png from dot

    This requires you install the node package [obographsviz](https://github.com/cmungall/obographviz)
    """
    def __init__(self,
                 image_format='png',
                 **args):
        super().__init__(**args)
        self.image_format = image_format
        # obographviz makes use of OboJson, so we leverage this here
        self.ojgr = OboJsonGraphRenderer(**args)

    # TODO: currently render and write are equivalent
    def render(self, ontol, query_ids=None, container_predicates=None, **args):
        if query_ids is None:
            query_ids = []
        if container_predicates is None:
            container_predicates = []
        g = ontol.get_graph()
        # create json object to pass to og2dot
        _, fn = tempfile.mkstemp(suffix='.json')
        self.ojgr.outfile = fn
        self.ojgr.write(ontol, **args)

        # call og2dot
        cmdtoks = ['og2dot.js']
        for q in query_ids:
            cmdtoks.append('-H')
            cmdtoks.append(q)
        cmdtoks.append('-t')
        cmdtoks.append(self.image_format)
        if len(container_predicates) > 0:
            for p in container_predicates:
                cmdtoks.append('-c')
                cmdtoks.append(p)
        if self.outfile is not None:
            cmdtoks.append('-o')
            cmdtoks.append(self.outfile)
        cmdtoks.append(fn)
        cp = subprocess.run(cmdtoks, check=True)
        logging.info(cp)
        os.remove(fn)
        
    def write(self, ontol, **args):
        self.render(ontol, **args)
            

class SimpleListGraphRenderer(GraphRenderer):
    """
    renders a graph as a simple flat list of nodes
    """
    def __init__(self, **args):
        super().__init__(**args)

    def render(self, ontol, **args):
        s = ""
        for n in ontol.nodes():
            s += self.render_noderef(ontol, n, **args) + "\n"
            #for n2 in ontol.parents(n):
            #    for _,ea in g.get_edge_data(n2,n).items():
            #        s += '  {} {}'.format(str(ea['pred']), self.render_noderef(ontol, n2, **args))
            #        s += "\n"
        return s
        
class AsciiTreeGraphRenderer(GraphRenderer):
    """
    Denormalized indented-text tree rendering
    """
    def __init__(self, **args):
        super().__init__(**args)
        
    def render(self, ontol, **args):
        #g = ontol.get_graph()
        #ts = nx.topological_sort(g)
        #roots = [n for n in ts if len(g.predecessors(n))==0]
        roots = ontol.get_roots()
        logging.info("Drawing ascii tree, using roots: {}".format(roots))
        if len(roots) == 0:
            logging.error("No roots in {}".format(ontol))
        s=""
        for n in roots:
            s += self._show_tree_node(None, n, ontol, 0, path=[], **args) + "\n"
        return s

    def _show_tree_node(self, rel, n, ontol, depth=0, path=None, **args):
        if path is None:
            path = []
        g = ontol.get_graph() # TODO - use ontol methods directly
        s = " " * depth + self.render_relation(rel) + " " +self.render_noderef(ontol, n, **args)
        if n in path:
            logging.warn("CYCLE: {} already visited in {}".format(n, path))
            return s + " <-- CYCLE\n"
        s += "\n"
        for c in ontol.children(n):
            preds = []
            for _,ea in g.get_edge_data(n,c).items():
                preds.append(ea['pred'])
            s+= self._show_tree_node(",".join(preds), c, ontol, depth+1, path+[n], **args)
        return s

class OboFormatGraphRenderer(GraphRenderer):
    """
    Render as obo format.

    Note this is currently incomplete
    """
    def __init__(self, **args):
        super().__init__(**args)
        
    def render(self, ontol, **args):
        ts = [n for n in ontol.nodes()]
        ts.sort()
        s = "ontology: auto\n\n"
        for n in ts:
            s += self.render_node(n, ontol, **args)
        return s
    
    def render_node(self, nid, ontol, **args):
        st = 'Term'
        if ontol.node_type == 'PROPERTY':
            st = 'Typedef'
            
        s = "[{}]\n".format(st)
        s += self.tag('id', nid)
        label = ontol.label(nid)
        if label is not None:
            s += self.tag('name', label)
            
        for p in ontol.parents(nid):
            for pred in ontol.child_parent_relations(nid,p):
                p_str = p
                if p in ontol.nodes():
                    p_label = ontol.label(p)
                    if p_label is not None:
                        p_str = '{} ! {}'.format(p, p_label)
                    
                if pred == 'subClassOf':
                    s += self.tag('is_a', p_str)
                else:
                    s += self.tag('relationship', pred, p_str)
        for ld in ontol.logical_definitions(nid):
            for gen in ld.genus_ids:
                s += self.tag('intersection_of', gen)
            for pred,filler in ld.restrictions:
                s += self.tag('intersection_of', pred, filler)

        td = ontol.text_definition(nid)
        if td:
            x = ""
            if td.xrefs is not None:
                x = ", ".join(td.xrefs)
            s += 'def: "{}" [{}]\n'.format(self._escape_quotes(td.val),x)
        else:
            logging.debug("No text def for: {}".format(nid))    

        logging.info("Looking up xrefs for {}".format(nid))
        for xref in ontol.xrefs(nid):
            s += "xref: {}\n".format(xref)

        if ontol.is_obsolete(nid):
            s += "is_obsolete: true\n"
            # TODO: consider links
            
        for syn in ontol.synonyms(nid):
            t = ""
            if syn.lextype is not None:
                t = " "+syn.lextype
            x = ""
            if syn.xrefs is not None:
                x = ", ".join(syn.xrefs)
            s += 'synonym: "{}" {}{} [{}]\n'.format(self._escape_quotes(syn.val),
                                                    syn.scope(),
                                                    t,
                                                    x)
        s += "\n"
        return s

    def _escape_quotes(self, v):
        # TODO: escape newlines etc
        return v.replace('"',"z").replace("\n", " ")
        
    # TODO
    def render_xrefs(self, nid, ontol, **args):
        g = ontol.xref_graph # TODO - use ontol methods directly
        n = g.node[nid]
        s = "[Term]\n"
        s += self.tag('id ! TODO', nid)
        s += self.tag('name', n['label'])
        for p in g.predecessors(nid):
            for _,ea in g.get_edge_data(p,nid).items():
                pred = ea['pred']
                if p in g and 'label' in g.node[p]:
                    p = '{} ! {}'.format(p, g.node[p]['label'])
                if pred == 'subClassOf':
                    s += self.tag('is_a', p)
                else:
                    s += self.tag('relationship', pred, p)
        s += "\n"
        return s
    
    def tag(self, t, *vs):
        v = " ".join(vs)
        return t + ': ' + v + "\n"

    # tag using dictionary
    def _dtag(self, t, d, k):
        if k in d:
            return self.tag(t, d[k])
        return ""
    
class OboJsonGraphRenderer(GraphRenderer):
    """
    Render as obographs json
    """
    def __init__(self, **args):
        super().__init__(**args)
        
    def to_json(self, ontol, **args):
        ontol.inline_xref_graph()
        g = ontol.get_graph() # TODO - use ontol methods directly
        obj = {}
        node_objs = []
        for n in ontol.nodes():
            node_objs.append(self.node_to_json(n, ontol, **args))
        obj['nodes'] = node_objs
        edge_objs = []
        for e in g.edges(data=True):
            edge_objs.append(self.edge_to_json(e, ontol, **args))
        obj['edges'] = edge_objs

        return {'graphs' : [obj]}

    def render(self, ontol, **args):
        obj = self.to_json(ontol, **args)
        return json.dumps(obj)
    
    def node_to_json(self, nid, ontol, **args):
        label = ontol.label(nid)
        if "include_meta" not in args or args['include_meta']:
            return {'id' : nid,
                    'lbl' : label,
                    'meta': ontol.node(nid).get('meta')}
        else:
            return {'id' : nid,
                'lbl' : label}
    
    def edge_to_json(self, e, ontol, **args):
        (obj,sub,meta) = e
        return {'sub' : sub,
                'obj' : obj,
                'pred' : meta['pred']}

    def tag(self, t, *vs):
        v = " ".join(vs)
        return t + ': ' + v + "\n"
    



import json

import typing
from typing import Dict, List, Optional

Message = Dict[str, str]

class Report(object):

    def __init__(self, group, dataset):
        self.group = group
        self.dataset = dataset
        self.messages = {} # type: Dict[str, List[Message]] # rule id --> List of messages
        self.messages["other"] = []
        self._rule_message_cap = 10000

    def _rule_id(self, id: int) -> str:
        """
        Convert an integer into a gorule key id.
        """
        if id is None or id == 0 or id >= 10000000:
            return "other"

        return "gorule-{:0>7}".format(id)

    def message(self, message: Message, rule: Optional[int]) -> None:
        """
        Add a message to the appropriate list of messages. If `rule` refers
        to a valid id range for a go rule, the message is entered in a list
        keyed by the full gorule-{id}. Otherwise, if `rule` is None, or
        outside the id range, then we put this in the catch-all "other"
        keyed list of messages.
        """
        rule_id = self._rule_id(rule)
        if rule_id not in self.messages:
            self.messages[rule_id] = []

        if len(self.messages[rule_id]) < self._rule_message_cap:
            self.messages[rule_id].append(message)

    def json(self, lines, associations, skipped) -> Dict:
        result = {
            "group": self.group,
            "dataset": self.dataset,
            "lines": lines,
            "skipped_lines": skipped,
            "associations": associations,
            "messages": self.messages
        }
        return result



from ontobio.io.entityparser import GpiParser
from ontobio.assoc_factory import AssociationSetFactory
from ontobio.ontol_factory import OntologyFactory

ONT = "tests/resources/go-truncated-pombase.json"
GPI = "tests/resources/truncated-mgi.gpi"

def test_parse_gpi():
    ont = OntologyFactory().create(ONT)
    p = GpiParser()
    p.config.remove_double_prefixes = True
    results = p.parse(open(GPI,"r"))
    for r in results:
        print(r)
    r1 = results[0]
    assert r1['label'] == 'a'
    assert r1['taxon']['id'] == 'NCBITaxon:10090'
    assert r1['full_name'] == 'nonagouti'
    assert r1['xrefs'] == ['UniProtKB:Q03288']
    assert r1['parents'] == []

    r2 = results[-1]
    assert r2['parents'] == ['MGI:109279']


    for m in p.report.messages:
        print("MESSAGE: {}".format(m))
    assert len(p.report.messages) == 0
    print(p.report.to_markdown())

def test_gpi_skips_lines_with_incorrect_collumns():
    parser = GpiParser()
    parsed = parser.parse_line("hello\tworld\tfoo\tbar")

    assert parsed[1] == []



import yaml
import os
import urllib

from typing import List
from dataclasses import dataclass

@dataclass
class Source:
    type: str
    url: str

@dataclass
class Dataset:
    name: str
    type: str
    sources: List[Source]

@dataclass
class Group:
    name: str



"""
Mapping between obograph-JSON format and networkx
"""

import itertools
import re
import json

import networkx
import logging
from prefixcommons.curie_util import contract_uri
from ontobio.ontol import LogicalDefinition
from ontobio.vocabulary.relations import map_legacy_pred

class OboJsonMapper(object):
    def __init__(self,
                 digraph=None,
                 context=None):
        self.digraph = digraph
        self.context = context if context is not None else {}

    def add_obograph_digraph(self, og, node_type=None, predicates=None, xref_graph=None, logical_definitions=None,
                             parse_meta=True,
                             **args):
        """
        Converts a single obograph to Digraph edges and adds to an existing networkx DiGraph
        """
        digraph = self.digraph
        logging.info("NODES: {}".format(len(og['nodes'])))

        # if client passes an xref_graph we must parse metadata
        if xref_graph is not None:
            parse_meta = True

        for n in og['nodes']:
            is_obsolete =  'is_obsolete' in n and n['is_obsolete'] == 'true'
            if is_obsolete:
                continue
            if node_type is not None and ('type' not in n or n['type'] != node_type):
                continue
            id = self.contract_uri(n['id'])
            digraph.add_node(id, **n)
            if 'lbl' in n:
                digraph.node[id]['label'] = n['lbl']
            if parse_meta and 'meta' in n:
                if n['meta'] is None:
                    n['meta'] = {}
                meta = self.transform_meta(n['meta'])
                if xref_graph is not None and 'xrefs' in meta:
                    for x in meta['xrefs']:
                        xref_graph.add_edge(self.contract_uri(x['val']), id, source=id)
        logging.info("EDGES: {}".format(len(og['edges'])))
        for e in og['edges']:
            sub = self.contract_uri(e['sub'])
            obj = self.contract_uri(e['obj'])
            pred = self.contract_uri(e['pred'])
            pred = map_legacy_pred(pred)
            if pred == 'is_a':
                pred = 'subClassOf'
            if predicates is None or pred in predicates:
                digraph.add_edge(obj, sub, pred=pred)
        if 'equivalentNodesSets' in og:
            nslist = og['equivalentNodesSets']
            logging.info("CLIQUES: {}".format(len(nslist)))
            for ns in nslist:
                equivNodeIds = ns['nodeIds']
                for i in ns['nodeIds']:
                    ix = self.contract_uri(i)
                    for j in ns['nodeIds']:
                        if i != j:
                            jx = self.contract_uri(j)
                            digraph.add_edge(ix, jx, pred='equivalentTo')
        if logical_definitions is not None and 'logicalDefinitionAxioms' in og:
            for a in og['logicalDefinitionAxioms']:
                ld = LogicalDefinition(self.contract_uri(a['definedClassId']),
                                       [self.contract_uri(x) for x in a['genusIds']],
                                       [(self.contract_uri(x['propertyId']),
                                         self.contract_uri(x['fillerId'])) for x in a['restrictions'] if x is not None])
                logical_definitions.append(ld)

    def transform_meta(self, m):
        if 'basicPropertyValues' in m:
            for x in m['basicPropertyValues']:
                x['pred'] = self.contract_uri(x['pred'])
                x['val'] = self.contract_uri(x['val'])
        return m

    def contract_uri(self, uri):
        if len(self.context.keys()) > 0:
            curies = contract_uri(uri, cmaps=[self.context])
            if len(curies) > 0:
                return sorted(curies, key=len)[0] # sort by length

        curies = sorted(contract_uri(uri), key=len) # Sort by length
        if len(curies) > 0:
            return curies[0]
        else:
            return uri


def convert_json_file(obographfile, **args):
    """
    Return a networkx MultiDiGraph of the ontologies
    serialized as a json string

    """
    f = open(obographfile, 'r')
    jsonstr = f.read()
    f.close()
    return convert_json_object(json.loads(jsonstr), **args)

def convert_json_object(obographdoc, **args):
    """
    Return a networkx MultiDiGraph of the ontologies
    serialized as a json object

    """
    digraph = networkx.MultiDiGraph()
    xref_graph = networkx.MultiGraph()
    logical_definitions = []
    context = obographdoc.get('@context',{})
    logging.info("CONTEXT: {}".format(context))
    mapper = OboJsonMapper(digraph=digraph, context=context)
    ogs = obographdoc['graphs']
    base_og = ogs[0]
    for og in ogs:
        # TODO: refactor this
        mapper.add_obograph_digraph(og, xref_graph=xref_graph,
                                    logical_definitions=logical_definitions, **args)

    return {
        'id': base_og.get('id'),
        'meta': base_og.get('meta'),
        'graph': mapper.digraph,
        'xref_graph': xref_graph,
        'graphdoc': obographdoc,
        'logical_definitions': logical_definitions
        }



from abc import ABCMeta, abstractmethod
from typing import Iterable, Optional, List, Dict
from ontobio.model.similarity import SimResult, IcStatistic
from ontobio.vocabulary.similarity import SimAlgorithm


class SimApi(metaclass=ABCMeta):

    @abstractmethod
    def search(self,
               id_list: Iterable,
               negated_classes: Iterable,
               limit: Optional[int],
               method: Optional) -> List[SimResult]:
        """
        Given an input list of classes, searches for similar lists of classes
        and provides a ranked list of matches
        """
        pass

    @abstractmethod
    def compare(self,
                query_classes: Iterable,
                reference_classes: Iterable,
                method: Optional) -> SimResult:
        """
        Given two lists of classes return their similarity
        """
        pass

    @staticmethod
    @abstractmethod
    def matchers() -> List[SimAlgorithm]:
        """
        Return the list of available matchers (eg resnik, boqa, cosine)
        could theoretically be a property with no setter
        """
        pass


class FilteredSearchable(metaclass=ABCMeta):

    @abstractmethod
    def filtered_search(self,
               id_list: Iterable,
               negated_classes: Iterable,
               limit: Optional[int],
               taxon_filter: Optional,
               category_filter: Optional,
               method: Optional) -> SimResult:
        """
        Given an input iterable of classes or individuals,
        provides a ranking of similar profiles
        """
        pass


class InformationContentStore(metaclass=ABCMeta):
    """
    Interface for an IC cache or store needed
    for various computations - eg annotation sufficiency
    """

    @property
    @abstractmethod
    def statistics(self):
        pass

    @statistics.setter
    @abstractmethod
    def statistics(self, value: IcStatistic):
        pass

    @property
    @abstractmethod
    def category_statistics(self):
        pass

    @category_statistics.setter
    @abstractmethod
    def category_statistics(self, value: Dict[str, IcStatistic]):
        pass

    @abstractmethod
    def get_profile_ic(self, profile: Iterable) -> Dict[str, float]:
        """
        Given a list of individuals,
        return a dictionary with their information content
        """
        pass



from ontobio.golr.golr_query import GolrSearchQuery
import pytest
import pysolr

"""
A group of integration tests that test integration
between a live solr instance and ontobio

Will fail if either monarch or go servers are down
"""

def test_search():
    q = GolrSearchQuery("abnormal")
    print("Q={}".format(q))
    params = q.solr_params()
    print("PARAMS={}".format(params))
    results = q.search()
    print("RESULTS={}".format(results))
    docs = results.docs
    for r in docs:
        print(str(r))
    assert len(docs) > 0


def test_solr_404():
    q = GolrSearchQuery("abnormal")
    q.update_solr_url("https://httpbin.org/status/404")
    pytest.raises(pysolr.SolrError, q.search)


def test_cursor():
    """
    Tests rows and start parameters.

    First fetch 100 docs, then same query but iterate with cursor in increments of ten.

    The two sets of IDs returned should be identical
    """
    q = GolrSearchQuery("abnormal", rows=100)
    results = q.search()
    docs = results.docs
    ids = set([d['id'] for d in docs])
    print('Init ids={}'.format(ids))
    assert len(ids) == 100
    matches = set()
    for i in range(0,10):
        q = GolrSearchQuery("abnormal", start=i*10, rows=10)
        docs = q.search().docs
        next_ids = [d['id'] for d in docs]
        assert len(next_ids) == 10
        print('Next ids (from {}) = {}'.format(i*10, next_ids))
        matches.update(next_ids)
    assert len(matches) == 100
    assert len(matches.intersection(ids)) == 100


def test_search_go_all():
    q = GolrSearchQuery("transport*", is_go=True)
    print("Q={}".format(q))
    params = q.solr_params()
    print("PARAMS={}".format(params))
    results = q.search()
    print("RESULTS={}".format(results))
    docs = results.docs
    for r in docs:
        print(str(r))
    assert len(docs) > 0
    print(str(results.facet_counts))


def test_search_go_ontol():
    q = GolrSearchQuery("transport*", category='ontology_class', is_go=True)
    print("Q={}".format(q))
    params = q.solr_params()
    print("PARAMS={}".format(params))
    results = q.search()
    print("RESULTS={}".format(results))
    docs = results.docs
    for doc in docs:
        print(str(doc))
    assert len(docs) > 0



import io

from ontobio.io import assocparser
from ontobio.io import gafparser
from ontobio.io import assocwriter

def test_gpad_iba_writing():
    out = io.StringIO()
    parser = gafparser.GafParser()
    parser.config = assocparser.AssocParserConfig(
        paint=True
    )
    writer = assocwriter.GpadWriter(file=out)

    for assoc in parser.association_generator(skipheader=True, file=open("tests/resources/wb_single_iba.gaf")):
        writer.write_assoc(assoc)

    outlines = out.getvalue().split("\n")

    expected_lines = [
        "!gpa-version: 1.1",
        "WB\tWBGene00022144\tpart_of\tGO:0005886\tPMID:21873635\tECO:0000318\tPANTHER:PTN000073732|RGD:3252\t\t20180308\tGO_Central\t\t",
        ""
    ]
    assert expected_lines == outlines
